# regression
from sklearn.metrics import (
    mean_squared_error,
    r2_score,
    explained_variance_score,
    median_absolute_error,
)

# classification
from sklearn.metrics import (
    accuracy_score,
    precision_score,
    recall_score,
    f1_score,
    roc_auc_score,
    balanced_accuracy_score,
)

import warnings
import pandas as pd
import numpy as np
from typing import Dict
import inspect

from model_monitoring.imputer import cluster_imputing
from model_monitoring.utils import (
    check_size,
    convert_Int_series,
    convert_Int_dataframe,
    convert_date_columns_to_seconds,
    convert_cluster_labels,
)
from model_monitoring.performance_measures.performance_measures import (
    compute_metric,
    compute_unsupervised_metric,
)
from model_monitoring.additional_metrics import classification_clustering


class PerformancesMeasures:
    """Performance Measures Class."""

    def __init__(
        self,
        approach_type="supervised",
        model_type="auto",
        set_metrics="standard",
        new_metrics=None,
        impute_nans=False,
        **kwargs,
    ):
        """Performance Measures Class.

        Args:
            approach_type (str): Type of approach: "supervised" or "unsupervised". Defaults to 'supervised'.
            model_type (str): Modelling problem among supervised ones: "regression", "classification", "multiclass" and "auto", or unsupervised one: "clustering". Defaults to 'auto'.
            set_metrics (str, optional): Metrics settings. It can be set "standard" for classical ML metrics, "new" for setting a new dictionary of metrics and "add" for adding new metrics to the standard ones. Defaults to 'standard'.
            new_metrics (dict/list, optional): If supervised approach: Dictionary of new metrics as keys when "set_metrics" is set to "new" and "add", and one among ['pred','prob'] as values. If unsupervised approach: list of new metrics. Defaults to None.
            impute_nans (bool): for unsupervised approach, if True, NaNs are filled if False they are not. Defaults to False

        """
        if approach_type not in ["supervised", "unsupervised"]:
            raise ValueError(
                f"{approach_type} is not a valid approach, select one of the following:\n['supervised','unsupervised']"
            )
        else:
            self.approach_type = approach_type
        # Check approach_type and model_type

        if self.approach_type == "supervised":
            if model_type in ["clustering"]:
                raise ValueError(
                    f"{model_type} is not a supervised model, select one of the following:\n['auto', 'regression', 'classification', 'multiclass'], or select an unsupervised one between:\n['clustering']"
                )
            elif model_type not in ["auto", "regression", "classification", "multiclass"]:
                raise ValueError(
                    f"{model_type} is not a valid algo_type. It should be one of the following:\n['auto', 'regression', 'classification', 'multiclass','clustering']"
                )
            else:
                self.model_type = model_type
                self.impute_nans = False
        else:
            if model_type in ["auto", "regression", "classification", "multiclass"]:
                raise ValueError(
                    f"{model_type} is not an unsupervised model, select one of the following:\n['clustering'], or select a supervised one between:\n['auto', 'regression', 'classification', 'multiclass']"
                )
            elif model_type not in ["clustering"]:
                raise ValueError(
                    f"{model_type} is not a valid algo_type. It should be one of the following:\n['auto', 'regression', 'classification', 'multiclass','clustering']"
                )
            else:
                self.model_type = model_type
                self.impute_nans = impute_nans
        # Check the set_metrics

        if set_metrics not in ["standard", "add", "new"]:
            raise ValueError(
                f"{set_metrics} is not a valid set_metrics. It should be one of the following:\n ['standard', 'add', 'new']"
            )
        self.set_metrics = set_metrics

        # Check new_metrics
        if self.set_metrics in ["add", "new"]:
            if self.approach_type == "supervised":
                if new_metrics is None:
                    self.new_metrics = {}
                else:
                    if isinstance(new_metrics, Dict):
                        if set(new_metrics.values()).issubset(set(["pred", "prob"])):
                            self.new_metrics = new_metrics
                        else:
                            raise ValueError(
                                f"{list(set(new_metrics.values()))} contains invalid input. Valid inputs are ['pred', 'prob']"
                            )
                    else:
                        raise ValueError(
                            f"{new_metrics} has not a valid format. It should be a dictionary containing functions as keys and one of ['pred', 'prob'] as values."
                        )
            else:
                if new_metrics is None:
                    self.new_metrics = []
                else:
                    if isinstance(new_metrics, list):
                        self.new_metrics = new_metrics
                    else:
                        raise ValueError(
                            f"{new_metrics} has not a valid format. It should be a list containing functions"
                        )
        else:
            self.new_metrics = new_metrics
        # Set the metrics for each set_metric case
        if self.set_metrics == "new":
            self.metrics = self.new_metrics
        if self.set_metrics in ["standard", "add"]:
            if self.model_type == "regression":
                self.metrics = {
                    mean_squared_error: "pred",
                    r2_score: "pred",
                    median_absolute_error: "pred",
                    explained_variance_score: "pred",
                }
            elif self.model_type == "classification":
                self.metrics = {
                    balanced_accuracy_score: "pred",
                    accuracy_score: "pred",
                    precision_score: "pred",
                    recall_score: "pred",
                    f1_score: "pred",
                    roc_auc_score: "prob",
                }

            elif self.model_type == "clustering":
                self.metrics = [classification_clustering]
            else:
                self.metrics = {balanced_accuracy_score: "pred", accuracy_score: "pred", roc_auc_score: "prob"}
            if self.set_metrics == "add":
                if self.approach_type == "supervised":
                    self.metrics = {**self.metrics, **self.new_metrics}
                else:
                    self.metrics = self.metrics + self.new_metrics

        # Check model_params
        self.predictions = None
        self.prob = None
        self.target = None
        self.perf_metrics = None
        self.cluster_labels = None
        self.data_matrix = None
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def compute_metrics(
        self,
        target=None,
        predictions=None,
        prob=None,
        cluster_labels=None,
        data_matrix=None,
        return_prop_true=True,
        classification_threshold=0.5,
        **kwargs,
    ):
        """Compute metrics performances.

        Args:
            target (np.array/pd.Series): target column
            predictions (np.array/pd.Series, optional): Predictions array. Defaults to None.
            prob (np.array/pd.Series, optional): Probabilities prob for classification. Defaults to None.
            cluster_labels (np.array/pd.Series, optional): Cluster labels array. Defaults to None
            data_matrix (dictionary/pd.Dataframe, optional): data to compute cluster metrics. Defaults to None
            return_prop_true (bool, optional): boolean that determines whether to return the portion of the target in binary classification. Defaults to True.
            classification_threshold (float, optional): threshold for binary classification predictions. Defaults to 0.5.

        Returns:
            dict: metrics performances
        """
        # Create perf_metric dict
        for k, v in kwargs.items():
            self.__dict__[k] = v
        perf_metrics = dict()
        # Metric if supervised approach
        if self.approach_type == "supervised":
            # Check one among predictions and prob exists
            if target is None:
                raise ValueError("target must not be None")
            if (predictions is None) and (prob is None):
                raise ValueError("at least one among predictions and prob must be not None")

            # Check size of the predictions and target
            if predictions is not None:
                check_size(predictions, target)

            # Check size of the target and prob
            if prob is not None:
                check_size(target, prob)

            if isinstance(target, np.ndarray):
                target = pd.Series(target, name="target")
            vals = target.nunique()
            if vals == 1:
                warnings.warn("The target column selected is constant")
            elif self.model_type == "auto":
                if vals <= 2:
                    self.model_type = "classification"
                elif vals < 11:
                    self.model_type = "multiclass"
                else:
                    self.model_type = "regression"

            if self.model_type == "regression":
                if predictions is None:
                    raise ValueError("predictions not provided")

            self.target = convert_Int_series(target)
            if predictions is not None:
                self.predictions = convert_Int_series(predictions)
            else:
                self.predictions = predictions
            if prob is not None:
                self.prob = convert_Int_series(prob)
            else:
                self.prob = prob

            self.return_prop_true = return_prop_true
            self.classification_threshold = classification_threshold

            if self.model_type == "regression":
                for i, j in self.metrics.items():
                    dict_to_use = {
                        k: v for k, v in self.__dict__.items() if k in inspect.signature(i).parameters.keys()
                    }
                    if j == "pred":
                        perf_metrics[i.__name__] = compute_metric(
                            self.target, self.predictions, metric=i, **dict_to_use
                        )
                    else:
                        warnings.warn(
                            f"{j} is a wrong label for regression model type. Label {i.__name__} with 'pred'."
                        )
            elif self.model_type == "classification":
                if self.predictions is None:
                    try:
                        self.predictions = self.prob.apply(lambda x: 1 if x > self.classification_threshold else 0)
                    except Exception:
                        self.predictions = np.array([1 if i > self.classification_threshold else 0 for i in self.prob])
                for i, j in self.metrics.items():
                    dict_to_use = {
                        k: v for k, v in self.__dict__.items() if k in inspect.signature(i).parameters.keys()
                    }
                    # some metrics don't work well with 1 target value (e.g. roc_auc_score)
                    if vals == 1:
                        if i.__name__ in ["roc_auc_score", "lift_score", "gain_score"]:
                            warnings.warn(f"{i.__name__} cannot be used when target has a constant value.")
                            continue
                    if j == "prob":
                        if self.prob is None:
                            warnings.warn(f"{i.__name__} needs prob, but prob are not provided")
                        else:
                            perf_metrics[i.__name__] = compute_metric(self.target, self.prob, metric=i, **dict_to_use)
                    else:
                        perf_metrics[i.__name__] = compute_metric(
                            self.target, self.predictions, metric=i, **dict_to_use
                        )
                if self.return_prop_true:
                    perf_metrics["proportion_1"] = (self.target == 1).mean()  # Add proportion of 1 label
            elif self.model_type == "multiclass":
                if self.predictions is None:
                    self.predictions = pd.Series([np.argmax(x) for x in self.prob], name="prob")
                for i, j in self.metrics.items():
                    dict_to_use = {
                        k: v for k, v in self.__dict__.items() if k in inspect.signature(i).parameters.keys()
                    }
                    if i.__name__ in [
                        "precision_score",
                        "recall_score",
                        "f1_score",
                        "fbeta_score",
                        "brier_score_loss",
                        "class_likelihood_ratios",
                        "dcg_score",
                        "jaccard_score",
                        "log_loss",
                        "matthews_corrcoef",
                        "ndcg_score",
                    ]:
                        warnings.warn(f"{i.__name__} is used for binary classification")
                    elif j == "prob":
                        if self.prob is None:
                            warnings.warn(f"{i.__name__} needs prob, but prob are not provided")
                        else:
                            perf_metrics[i.__name__] = compute_metric(
                                self.target, self.prob, metric=i, multi_class="ovr"
                            )
                    else:
                        perf_metrics[i.__name__] = compute_metric(
                            self.target, self.predictions, metric=i, **dict_to_use
                        )
            else:
                raise ValueError("Invalid model type.")
        # Metric if unsupervised approach
        else:
            # data quality
            if (cluster_labels is None) or (data_matrix is None):
                raise ValueError("Both cluster_labels and data_matrix must be not None")
            else:
                check_size(data_matrix, cluster_labels)
                data_matrix = convert_date_columns_to_seconds(data_matrix)
            if not isinstance(cluster_labels, pd.Series):
                cluster_labels = pd.Series(cluster_labels, name="cluster_labels")
            cluster_labels = convert_Int_series(cluster_labels)

            if cluster_labels.isna().any():
                condition = ~cluster_labels.isna()
                cluster_labels = cluster_labels[condition].reset_index(drop=True)
                data_matrix = data_matrix[condition].reset_index(drop=True)

            vals = cluster_labels.nunique()
            # uniform cluster value and type
            cluster_labels = convert_cluster_labels(cluster_labels, vals)

            if not isinstance(data_matrix, pd.DataFrame):
                data_matrix = pd.DataFrame(data_matrix)
            data_matrix = convert_Int_dataframe(data_matrix)
            if data_matrix.isna().any().any():
                if self.impute_nans:
                    valid_params = inspect.signature(cluster_imputing).parameters.keys()
                    dict_to_use = {k: v for k, v in self.__dict__.items() if k in valid_params}

                    self.data_matrix, self.cluster_labels = cluster_imputing(data_matrix, cluster_labels, **dict_to_use)
                else:
                    self.data_matrix = data_matrix
                    self.cluster_labels = cluster_labels
            else:
                self.data_matrix = data_matrix
                self.cluster_labels = cluster_labels

            categorical_cols = self.data_matrix.select_dtypes(exclude=["number"]).columns
            if self.new_metrics:
                data_matrix_dummies = self.data_matrix.copy()
                data_matrix_dummies = pd.get_dummies(data_matrix_dummies, columns=categorical_cols)
            self.data_matrix[categorical_cols] = self.data_matrix[categorical_cols].astype("category")
            # warning for classification clustering
            # compute metrics
            if vals != 1:
                for i in self.metrics:
                    dict_to_use = {
                        k: v for k, v in self.__dict__.items() if k in inspect.signature(i).parameters.keys()
                    }
                    try:
                        if i.__name__ != "classification_clustering":
                            metric = compute_unsupervised_metric(
                                data_matrix_dummies, self.cluster_labels, i, **dict_to_use
                            )
                            perf_metrics[i.__name__] = metric
                        else:
                            metric = compute_unsupervised_metric(
                                self.data_matrix, self.cluster_labels, i, **dict_to_use
                            )
                            perf_metrics[i.__name__] = metric
                    except Exception:
                        if not self.impute_nans:
                            if self.data_matrix.isna().any().any():
                                raise ValueError(
                                    f"\n An error occured while computing {i.__name__}, the reason could be impute_nans is set to False, try set it as True while initializating the class"
                                )
                        else:
                            raise ValueError(
                                f"\n{i.__name__} is not a valid additional metric for clustering monitoring"
                            )
            # Many metrics don't work well with a single cluster
            else:
                for i in self.metrics:
                    try:
                        if i.__name__ != "classification_clustering":
                            metric = compute_unsupervised_metric(
                                data_matrix_dummies, self.cluster_labels, i, **dict_to_use
                            )
                            perf_metrics[i.__name__] = metric
                        else:
                            metric = compute_unsupervised_metric(
                                self.data_matrix, self.cluster_labels, i, **dict_to_use
                            )
                            perf_metrics[i.__name__] = metric
                    except Exception:
                        warnings.warn(
                            f"\n{i.__name__} is not a valid additional metric for clustering monitoring with a single cluster"
                        )
                        pass
        self.perf_metrics = perf_metrics
        return self.perf_metrics



import pandas as pd
import numpy as np
from sklearn.metrics import recall_score, balanced_accuracy_score
from sklearn.model_selection import StratifiedKFold, KFold
from model_monitoring.performance_measures import compute_unsupervised_metric
import lightgbm as lgb


def lift_score(y_true, y_prob, method: str = "percentile", percentile: float = 0.2, threshold: float = 0.5):
    """Calculates the lift score which compares the model true positive rate versus a random choice.

    Args:
        y_true (np.array, pd.Series): real value of target
        y_prob (np.array, pd.Series): predicted probability of the target
        method (str, optional): 'percentile' or 'threshold'. If percentile it uses the defined percentiles to set the cutoff.
            If threshold it uses the defined threshold as cutoff. Default percentile
        percentile (float, optional): Percentile to use to get the cutoff. Defaults to 0.2.
        threshold (float, optional): Threshold to use as cutoff. Defaults to 0.5.

    Returns:
        float: lift score
    """
    checker_te = pd.DataFrame({"events": y_true, "prob": y_prob})
    if method == "percentile":
        cutoff = checker_te.prob.quantile(q=1 - percentile)
    else:
        cutoff = threshold
    if checker_te["events"].sum() == 0:
        raise ValueError("y_true always 0")

    tpr_db = checker_te.loc[checker_te.prob > cutoff]
    tpr_model = tpr_db["events"].sum() / tpr_db.shape[0]
    tpr_random = checker_te["events"].sum() / checker_te.shape[0]

    return tpr_model / tpr_random


def gain_score(y_true, y_prob, method: str = "percentile", percentile: float = 0.2, threshold: float = 0.5):
    """Calculates the gain score which identifies the recall at a cutoff.

    Args:
        y_true (np.array, pd.Series): real value of target
        y_prob (np.array, pd.Series): predicted probability of the target
        method (str, optional): 'percentile' or 'threshold'. If percentile it uses the defined percentiles to set the cutoff.
            If threshold it uses the defined threshold as cutoff. Default percentile
        percentile (float, optional): Percentile to use to get the cutoff. Defaults to 0.2.
        threshold (float, optional): Threshold to use as cutoff. Defaults to 0.5.

    Returns:
        float: gain score
    """
    checker_te = pd.DataFrame({"events": y_true, "prob": y_prob})
    if method == "percentile":
        cutoff = checker_te.prob.quantile(q=1 - percentile)
    else:
        cutoff = threshold
    if checker_te["events"].sum() == 0:
        raise ValueError("y_true always 0")
    return recall_score(y_true, np.where(checker_te.prob > cutoff, 1, 0))


def classification_clustering(
    data, labels, n_splits=5, classification_clustering_metric=balanced_accuracy_score, **kwargs
):
    """Train model for classification clustering method and compute classification clustering metric.

    Args:
        data: (pd.Dataframe): original data matrix
        labels: (pd.series/np.array): cluster labels
        test_size (float, optional): ratio between original dataset and test set dimensions. Defaults to 0.3
        classification_clustering_metric (func, optional): classification clustering metric to compute

    Returns:
        float: classification_clustering_metric
    """
    model = lgb.LGBMClassifier(class_weight="balanced")
    kf = KFold(n_splits=n_splits, shuffle=True)
    skf = StratifiedKFold(n_splits=n_splits, shuffle=True)
    lst_perf = []
    try:
        for train_index, test_index in skf.split(data, labels):
            x_train_fold = data.loc[train_index]
            x_test_fold = data.loc[test_index]
            y_train_fold, y_test_fold = labels[train_index], labels[test_index]
            model.fit(x_train_fold, y_train_fold)
            predicted_labels = model.predict(x_test_fold)
            metric = compute_unsupervised_metric(
                y_test_fold, predicted_labels, classification_clustering_metric, **kwargs
            )
            lst_perf.append(metric)
    except Exception:
        for train_index, test_index in kf.split(data, labels):
            x_train_fold = data.loc[train_index]
            x_test_fold = data.loc[test_index]
            y_train_fold, y_test_fold = labels[train_index], labels[test_index]
            model.fit(x_train_fold, y_train_fold)
            predicted_labels = model.predict(x_test_fold)
            metric = compute_unsupervised_metric(
                y_test_fold, predicted_labels, classification_clustering_metric, **kwargs
            )
            lst_perf.append(metric)
    return np.median(lst_perf)



import pandas as pd
import numpy as np
import numba
from numba import njit


def convert_columns_name(column_names):
    """Converts a list of column names to a mapping of names to numbers and vice versa.

    Args:
        column_names (list of str): A list of column names to be converted.

    Returns:
        (dict): A dictionary mapping each column name to a unique integer.
        (dict): A dictionary mapping each integer back to the original column name.
    """
    numbers = {name: str(i) for i, name in enumerate(column_names)}
    reverse_mapping = {v: k for k, v in numbers.items()}
    return numbers, reverse_mapping


def rename_columns_in_dataframe(df, column_mapping):
    """
    Renames columns in a DataFrame according to a specified mapping. The keys are the existing column names and the values are the new column names.

    Args:
        df (pandas.DataFrame): The DataFrame whose columns are to be renamed.
        column_mapping (dict): A dictionary mapping current column names to new column names.

    Returns:
        None: The function modifies the DataFrame in place.
    """
    df.rename(columns=column_mapping, inplace=True)


def cluster_imputing(data, labels, k=5, size=500):
    """Performs K-Nearest Neighbors (KNN) imputation on the input dataset, grouping by provided labels.

    Args:
        data (pd.DataFrame): The input dataset containing missing values.
        labels (pd.Series): A series used to group the data for imputation.
        k (int, optional): The number of nearest neighbors to consider for imputation. Defaults to 5.
        size (int, optional): The size of sample in which searching the k-neighbours. Defaults to 500.

    Returns:
        pd.DataFrame: The imputed dataset with missing values filled.
        np.ndarray: An array of cluster labels corresponding to the input data.
    """
    random_strings_dict, reverse_mapping = convert_columns_name(data.columns)
    rename_columns_in_dataframe(data, random_strings_dict)
    data_grouped = knn__dataset_prep(data, labels)

    imputed_data = knn_imputation(data, data_grouped, k, size)
    cluster_labels = np.asarray(imputed_data["cluster"].values)
    imputed_data = imputed_data.drop(columns=["cluster", "original_index_to_drop"])
    rename_columns_in_dataframe(imputed_data, reverse_mapping)
    return imputed_data, cluster_labels


def knn__dataset_prep(data, labels):
    """Applies KNN imputation on grouped data, handling missing values within each group.

    Args:
        data (pd.DataFrame): The original dataset containing missing values.
        labels (np.array): Dataset cluster labels.

    Returns:
        pd.core.groupby.generic.DataFrameGroupBy: The grouped dataset for imputation.
    """
    datanan = data.index[data.isnull().all(axis=1)]
    data["cluster"] = labels
    data["cluster"] = data["cluster"].astype(str)
    data = data.drop(datanan).reset_index(drop=True)
    data["original_index_to_drop"] = data.index.astype(int)

    data_grouped = data.groupby("cluster")
    return data_grouped


def knn_imputation(data, data_grouped, k=5, size=500):
    """Applies KNN imputation on grouped data, handling missing values within each group.

    Args:
        data (pd.DataFrame): The original dataset containing missing values.
        data_grouped (pd.core.groupby.generic.DataFrameGroupBy): The grouped dataset for imputation.
        k (int, optional): The number of nearest neighbors to consider for imputation. Defaults to 5.
        size (int, optional): The size of sample in which searching the k-neighbours. Defaults to 500.

    Returns:
        pd.DataFrame: The dataset with missing values imputed.
    """
    imputed_data = []
    for a, group in data_grouped:
        copy_group = group.copy()
        group = group.drop(columns=["cluster", "original_index_to_drop"])
        if group.isna().all().any():
            for col in group:
                if group[col].isna().all():
                    if not pd.api.types.is_numeric_dtype(group[col]):
                        mode = data[col].mode()[0] if not data[col].mode().empty else None
                        group[col].fillna(mode, inplace=True)
                    else:
                        group[col].fillna(data[col].median(), inplace=True)
        group = imputer_numba(group, k, size)
        group["cluster"] = copy_group["cluster"]
        group["original_index_to_drop"] = copy_group["original_index_to_drop"]
        imputed_data.append(group)

    imputed_data = pd.concat(imputed_data)
    imputed_data = imputed_data.sort_values("original_index_to_drop")
    return imputed_data


def data_prep(df):
    """
    Prepares the data by converting categorical columns into dummy variables and handling NaN values.

    Args:
        df (pandas.DataFrame): The original DataFrame containing categorical data.

    Returns:
        df_dummies (pandas.DataFrame): The DataFrame with dummy columns added.
        original_categorical_columns (list): A list of the original categorical columns.
    """
    df_dummies = pd.get_dummies(df, dummy_na=False)
    original_categorical_columns = []
    for col in df.columns:
        dummy_cols = [dcol for dcol in df_dummies.columns if dcol.startswith(col + "_") and dcol not in df.columns]
        if dummy_cols:
            original_categorical_columns.append(col)
    nan_mask = df[original_categorical_columns].isna()

    for col in original_categorical_columns:
        dummy_cols = [col + "_" + str(val) for val in df[col].dropna().unique()]
        df_dummies[dummy_cols] = np.where(nan_mask[col].values[:, None], np.nan, df_dummies[dummy_cols].values)

    return df_dummies, original_categorical_columns


def revert_dummies(df_imputed, df, original_categorical_columns):
    """
    Converts dummy-encoded columns back to their original categorical format.

    Args:
        df_imputed (pandas.DataFrame): The DataFrame with dummy columns after imputation.
        df (pandas.DataFrame): The original DataFrame before dummy encoding.
        original_categorical_columns (list): A list of the original categorical columns.

    Returns:
        pandas.DataFrame: The DataFrame with the original categorical columns restored.
    """
    original_order = df.columns.tolist()
    for prefix in original_categorical_columns:
        dummy_cols = [col for col in df_imputed.columns if col.startswith(prefix + "_")]
        if dummy_cols:
            df_imputed[prefix] = df_imputed[dummy_cols].idxmax(axis=1)
            df_imputed[prefix] = df_imputed[prefix].apply(
                lambda x: x[len(prefix) + 1 :] if pd.notna(x) and "_" in x else np.nan
            )
            df_imputed = df_imputed.drop(columns=dummy_cols)

    df_imputed = df_imputed[original_order]

    return df_imputed


@numba.vectorize
def calc_mono_dist(val1, val2, max_val, min_val):
    """
    Calculates the normalized distance between two values, handling NaN values.

    Args:
        val1 (float): The first value.
        val2 (float): The second value.
        max_val (float): The maximum value of the variable.
        min_val (float): The minimum value of the variable.

    Returns:
        float: distance between two values.
    """
    if np.isnan(val1) or np.isnan(val2):
        return 1
    return abs(val1 - val2) / (max_val - min_val) if max_val != min_val else 0


@numba.jit(nopython=True)
def custom_distance_per_variable(point1, point2, max_vals, min_vals):
    """
    Computes the custom distance between two points across multiple variables.

    Args:
        point1 (np.ndarray): The first point as an array of values.
        point2 (np.ndarray): The second point as an array of values.
        max_vals (np.ndarray): Array of maximum values for each variable.
        min_vals (np.ndarray): Array of minimum values for each variable.

    Returns:
        float: The mean of the calculated distances across all variables.
    """
    distances = np.zeros(len(point1))
    for i in range(len(point1)):
        distances[i] = calc_mono_dist(point1[i], point2[i], max_vals[i], min_vals[i])
    return distances


@numba.jit(nopython=True)
def calculate_distance_for_point(k_point, random_points, max_vals, min_vals):
    """
    Compute distances between a point and an array of points.

    Args:
        punto_k (np.ndarray): Point of which distances will be calculated.
        punti_casuali (np.ndarray): Random point matrix.
        max_vals (np.ndarray): Maximum values array.
        min_vals (np.ndarray): Minimum values array.

    Returns:
        np.ndarray: Distance array between a point and each sampled points of dataset.
    """
    n_points = random_points.shape[0]
    n_variables = k_point.shape[0]

    distances = np.empty((n_points, n_variables))

    for i in range(n_points):
        distances[i, :] = custom_distance_per_variable(k_point, random_points[i], max_vals, min_vals)

    return distances


@njit
def fix_std(std_dev):
    """
    Changes std value to 1 if the real value is 0.

    Args:
        std_dev (np.ndarray): array of std of a set of variables.

    Return:
        np.ndarray: array of fixed std's.
    """
    fixed_std_dev = std_dev.copy()
    for i in range(len(fixed_std_dev)):
        if fixed_std_dev[i] == 0:
            fixed_std_dev[i] = 1
    return fixed_std_dev


@numba.njit
def standardize_distances(distances, mean_dist, std_dist):
    """
    Standardizes distances using the mean and standard deviation.

    Args:
        distances (np.ndarray): Array of distances to standardize.
        mean_dist (float): Mean of the distances.
        std_dist (float): Standard deviation of the distances.

    Returns:
        np.ndarray: Array of standardized distances.
    """
    standardized_distances = np.empty_like(distances)
    std_dist = fix_std(std_dist)
    for i in range(distances.shape[0]):
        standardized_distances[i] = (distances[i] - mean_dist) / std_dist
    return standardized_distances


@numba.njit
def row_means(matrix):
    """
    Computes the mean of each row in a matrix.

    Args:
        matrix (np.ndarray): 2D array from which to compute row means.

    Returns:
        np.ndarray: 1D array of row means.
    """
    n_rows = matrix.shape[0]
    row_means = np.empty(n_rows)

    for i in range(n_rows):
        row_means[i] = np.mean(matrix[i, :])

    return row_means


def create_distance_array(index, df_values, max_vals, min_vals, size=500):
    """
    Creates an array of standardized distances for a specific point in the DataFrame.

    Args:
        index (int): Index of the point in the DataFrame.
        df_values (np.ndarray): Array of DataFrame values.
        max_vals (np.ndarray): Array of maximum values for each variable.
        min_vals (np.ndarray): Array of minimum values for each variable.
        size (int, optional): Number of random points to sample (default is 500).

    Returns:
        np.ndarray: Array of standardized distances.
        np.ndarray: Array of sampled points.
    """
    k_point = df_values[index]
    if df_values.shape[0] > 5000:
        random_indices = np.random.choice(df_values.shape[0], size=size, replace=False)
    else:
        random_indices = np.arange(df_values.shape[0])
    random_points = df_values[random_indices]

    distances = calculate_distance_for_point(k_point, random_points, max_vals, min_vals)
    mean_dist = np.nanmean(distances, axis=0)
    std_dist = np.nanstd(distances, axis=0)
    standardized_distances = standardize_distances(distances, mean_dist, std_dist)
    standardized_distances = row_means(standardized_distances)
    return standardized_distances, random_points


@numba.jit(nopython=True)
def find_k_nearest(distances, k):
    """
    Finds the indices of the k-nearest points based on the distances.

    Args:
        distances (np.ndarray): Array of distances.
        k (int): Number of nearest points to find.

    Returns:
        np.ndarray: Indices of the k-nearest points.
    """
    return np.argsort(distances)[:k]


@numba.jit(nopython=True)
def impute_missing_values(row, nearest_rows, df_mean):
    """
    Imputes missing values in a row using the mean of the nearest rows.

    Args:
        row (np.ndarray): The row with missing values.
        nearest_rows (np.ndarray): Array of nearest rows.
        df_mean (np.ndarray): Mean values for each column in the DataFrame.

    Returns:
        np.ndarray: The row with imputed values.
    """
    for i in range(len(row)):
        if np.isnan(row[i]):
            row[i] = np.nanmean(nearest_rows[:, i])
        if np.isnan(row[i]):
            row[i] = df_mean[i]
    return row


def imputer_numba(df, k=5, size=500):
    """
    Imputes missing values in a DataFrame using a custom distance metric and Numba for optimization.

    Args:
        df (pd.DataFrame): The DataFrame with missing values.
        k (int, optional): The number of nearest neighbors to consider for imputation. Defaults to 5.
        size (int, optional): The size of sample in which searching the k-neighbours. Defaults to 500.

    Returns:
        pd.Dataframe: the imputed DataFrame
    """
    df2 = df.copy()
    df = df.dropna(how="all")
    df, original_categorical = data_prep(df)
    max_vals = df.max().values
    min_vals = df.min().values
    df_mean = np.nanmean(df.values, axis=0)
    nan_mask = df.isna().any(axis=1).values
    nan_row_indices = np.where(nan_mask)[0]
    df_values = df.values
    # imputation
    for index in nan_row_indices:
        distance_array, sample_array = create_distance_array(index, df_values, max_vals, min_vals, size)
        k_nearest = find_k_nearest(distance_array, k)
        df_values[index] = impute_missing_values(df_values[index], sample_array[k_nearest], df_mean)
    df = pd.DataFrame(df_values, columns=df.columns, index=df.index)
    # revert filled dataset
    df = revert_dummies(df, df2, original_categorical)

    return df
